@@ -1,6 +1,7 @@
 import sys
 import asyncio
-import heapq
+import asyncio.events as events
+import threading
 
 
 def apply(loop=None):
@@ -10,7 +11,7 @@ def apply(loop=None):
     loop = loop or asyncio.get_event_loop()
     if not isinstance(loop, asyncio.BaseEventLoop):
         raise ValueError('Can\'t patch loop of type %s' % type(loop))
-    if hasattr(loop, '_run_until_complete_orig'):
+    if hasattr(loop, '_run_forever_orig'):
         # already patched
         return
     _patch_asyncio()
@@ -26,12 +27,8 @@ def _patch_asyncio():
     """
     def run(future, *, debug=False):
         loop = asyncio.get_event_loop()
-        run_orig = asyncio._run_orig  # noqa
-        if run_orig and not loop.is_running():
-            return run_orig(future, debug=debug)
-        else:
-            loop.set_debug(debug)
-            return loop.run_until_complete(future)
+        loop.set_debug(debug)
+        return loop.run_until_complete(future)
 
     if sys.version_info >= (3, 6, 0):
         asyncio.Task = asyncio.tasks._CTask = asyncio.tasks.Task = \
@@ -50,74 +47,48 @@ def _patch_loop(loop):
     """
     Patch loop to make it reentrent.
     """
-    def run_until_complete(self, future):
-        if self.is_running():
-            self._check_closed()
-            f = asyncio.ensure_future(future)
-            if f is not future:
-                f._log_destroy_pending = False
-            while not f.done():
-                run_once(self)
-            return f.result()
-        else:
-            return self._run_until_complete_orig(future)
+    if sys.version_info >= (3, 7, 0):
+        set_coro_tracking = loop._set_coroutine_origin_tracking
+    else:
+        set_coro_tracking = loop._set_coroutine_wrapper
 
-    bogus_handle = asyncio.Handle(None, None, loop)
-    bogus_handle.cancel()
+    def run_forever(self):
+        self._check_closed()
+        old_thread_id = self._thread_id
+        old_running_loop = events._get_running_loop()
+        set_coro_tracking(self._debug)
+        self._thread_id = threading.get_ident()
 
-    def run_once(self):
-        ready = self._ready
-        scheduled = self._scheduled
-
-        # remove bogus handles to get more efficient timeout
-        while ready and ready[0] is bogus_handle:
-            ready.popleft()
-        nready = len(ready)
-
-        while scheduled and scheduled[0]._cancelled:
-            self._timer_cancelled_count -= 1
-            handle = heapq.heappop(scheduled)
-            handle._scheduled = False
-
-        timeout = None
-        if ready or self._stopping:
-            timeout = 0
-        elif scheduled:
-            when = scheduled[0]._when
-            timeout = max(0, when - self.time())
-
-        event_list = self._selector.select(timeout)
-        self._process_events(event_list)
-
-        end_time = self.time() + self._clock_resolution
-        while scheduled:
-            handle = scheduled[0]
-            if handle._when >= end_time:
-                break
-            handle = heapq.heappop(scheduled)
-            handle._scheduled = False
-            ready.append(handle)
-
-        self._nesting_level += 1
-        ntodo = len(ready)
-        for _ in range(ntodo):
-            if not ready:
-                break
-            handle = ready.popleft()
-            if handle._cancelled:
-                continue
-            handle._run()
-        handle = None
-        self._nesting_level -= 1
-
-        # add bogus handles to keep loop._run_once happy
-        if nready and self._nesting_level == 0:
-            ready.extendleft([bogus_handle] * nready)
+        old_agen_hooks = sys.get_asyncgen_hooks()
+        if self._asyncgens is not None:
+            sys.set_asyncgen_hooks(
+                firstiter=self._asyncgen_firstiter_hook,
+                finalizer=self._asyncgen_finalizer_hook)
+        try:
+            events._set_running_loop(self)
+            while True:
+                try:
+                    self._run_once()
+                except IndexError:
+                    # Ignore 'pop from an empty deque' errors.
+                    # This happens when all ready handles have already been
+                    # processed but _run_once expects there to be more.
+                    # Since the handles have been processed anyway it is
+                    # safe to ignore this.
+                    pass
+                if self._stopping:
+                    break
+        finally:
+            self._stopping = False
+            self._thread_id = old_thread_id
+            events._set_running_loop(old_running_loop)
+            set_coro_tracking(False)
+            if self._asyncgens is not None:
+                sys.set_asyncgen_hooks(*old_agen_hooks)
 
     cls = loop.__class__
-    cls._run_until_complete_orig = cls.run_until_complete
-    cls.run_until_complete = run_until_complete
-    cls._nesting_level = 0
+    cls._run_forever_orig = cls.run_forever
+    cls.run_forever = run_forever
 
 
 def _patch_task():
